/*
 * Copyright (c) 1991,1993 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: /usr/src/mash/repository/vint/nam-1/edge.cc,v 1.9 1997/11/22 22:59:26 tecklee Exp $ (LBL)";
#endif

#include "config.h"
#include "sincos.h"

#include "edge.h"
#include "node.h"
#include "packet.h"
#include "view.h"
#include "netview.h"
#include "transform.h"
#include "paint.h"
#include "monitor.h"

Edge::Edge(Node* src, Node* dst, double ps,
	   double bw, double delay, double angle) :
	Animation(0, 0),
	src_(src->num()), dst_(dst->num()),
	neighbor_(dst),
	x0_(0), y0_(0),
	x1_(0), y1_(0),
	psize_(ps),
	angle_(angle),
	bandwidth_(bw),
	delay_(delay),
	state_(UP),
	packets_(NULL),
	no_of_packets_(0),
	last_packet_(NULL),
	marked_(0)
{
	paint_ = Paint::instance()->thick();
}

Edge::~Edge()
{
}

void Edge::init_color(char *clr)
{
	color(clr);
	oldPaint_ = paint_;
}

void Edge::set_down(char *color)
{
	// If current color is down, don't change it again.
	// Assuming only one down color. User can change this behavior
	// by adding tcl code for link-up and link-down events.
	if (state_ == UP) {
		int pno = Paint::instance()->lookup(Tk_GetUid(color), 3);
		oldPaint_ = paint_;
		paint_ = pno;
		state_ = DOWN;
	}
}

void Edge::set_up()
{
	if (state_ == DOWN) {
		state_ = UP;
		toggle_color();
	}
}

void Edge::place(double x0, double y0, double x1, double y1)
{
	x0_ = x0;
	y0_ = y0;
	x1_ = x1;
	y1_ = y1;

	double dx = x1 - x0;
	double dy = y1 - y0;
	/*XXX*/
	// delay should not be equal to edge's position
	//	delay_ = sqrt(dx * dx + dy * dy);
	matrix_.clear();
	matrix_.rotate((180 / M_PI) * atan2(dy, dx));
	matrix_.translate(x0, y0);

	if (x1>x0) {
	  bb_.xmin = x0; bb_.xmax = x1;
	} else {
	  bb_.xmin = x1; bb_.xmax = x0;
	}
	if (y1>y0) {
	  bb_.ymin = y0; bb_.ymax = y1;
	} else {
	  bb_.ymin = y1; bb_.ymax = y0;
	}
}

void Edge::AddPacket(Packet *p)
{
  p->next(packets_);
  if (packets_!=NULL)
    packets_->prev(p);
  p->prev(NULL);
  packets_=p;
  if (last_packet_==NULL)
    last_packet_=p;
  no_of_packets_++;
#define PARANOID
#ifdef PARANOID
  int ctr=0;
  for(Packet *tp=packets_;tp!=NULL;tp=tp->next())
    {
      ctr++;
      if (ctr>no_of_packets_) abort();
    }
  if (last_packet_->next()!=NULL) abort();
#ifdef DEBUG
  printf("AddPacket: %d->%d OK\n", src_, dst_);
#endif
#endif
  
}

void Edge::DeletePacket(Packet *p)
{
  no_of_packets_--;
  if (last_packet_==p)
    {
      /*Normal case - it's the last packet*/
      last_packet_ = p->prev();
      if (packets_!=p)
	p->prev()->next(NULL);
      else
	packets_=NULL;
    }
  else
    {
      /*Abnormal case - need to search the list*/
      Packet *tp;
      tp=packets_;
      while((tp!=p)&&(tp!=NULL))
	{
	  tp=tp->next();
	}
      if (tp==p)
	{
	  if (tp==packets_)
	    /*it was the first packet*/
	    packets_=tp->next();
	  else
	    tp->prev()->next(tp->next());
	  if (tp==last_packet_)
	    {
	      /*this shouldn't ever happen*/
	      fprintf(stderr, "**it happened\n");
	      last_packet_=tp->prev();
	    }
	  else
	    tp->next()->prev(tp->prev());
	}
    }
#define PARANOID
#ifdef PARANOID
  int ctr=0;
  for(Packet *tp=packets_;tp!=NULL;tp=tp->next())
    {
      ctr++;
      if (ctr>no_of_packets_) abort();
    }
  if ((last_packet_!=NULL)&&(last_packet_->next()!=NULL)) abort();
#ifdef DEBUG
  printf("DeletePacket: %d->%d OK\n", src_, dst_);
#endif
#endif
}

void Edge::draw(NetView* view, double /*now*/) const
{
	view->line(x0_, y0_, x1_, y1_, paint_);
	if (monitor_!=NULL)
	  monitor_->draw(view, (x0_+x1_)/2, (y0_+y1_)/2);
}

void Edge::reset(double)
{
	//	paint_ = Paint::instance()->thick();
	no_of_packets_=0;
	packets_=NULL;
	last_packet_=NULL;
}

int Edge::inside(double, float px, float py) const
{
  return (px >= bb_.xmin &&
		px <= bb_.xmax &&
		py >= bb_.ymin - .0005 &&
		py <= bb_.ymax + .0005);
}

const char* Edge::info() const
{
	static char text[128];
	sprintf(text, "link %d-%d:\n  bw: %g bits/sec\n  delay: %g sec\n",
		src_, dst_, bandwidth_, delay_);
	return (text);
}

const char* Edge::getname() const
{
	static char text[128];
	sprintf(text, "l %d %d",
		src_, dst_);
	return (text);
}

void Edge::monitor(Monitor *m, double /*now*/, char *result, int /*len*/)
{
  monitor_=m;
  sprintf(result, "link %d-%d:\n bw: %g bits/sec\n delay: %g sec",
	  src_, dst_, bandwidth_, delay_);
  return;
}
