/*
 * Copyright (c) 1991,1993 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Computer Systems
 *      Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
"@(#) $Header: /usr/src/mash/repository/vint/nam-1/netmodel.cc,v 1.18 1997/12/16 06:42:24 heideman Exp $ (LBL)";
#endif

#include <stdlib.h>
#include <ctype.h>
#include <math.h>

#include "netview.h"
#include "animation.h"
#include "queue.h"
#include "drop.h"
#include "packet.h"
#include "edge.h"
#include "lan.h"
#include "node.h"
#include "agent.h"
#include "feature.h"
#include "route.h"
#include "netmodel.h"
#include "monitor.h"
#include "trace.h"
#include "paint.h"
#include "sincos.h"
#include "state.h"

#include <float.h>

extern int lineno;
// not used
//static int next_pat;

/*XXX */
double defsize = 0.;

class NetworkModelClass : public TclClass {
 public:
	NetworkModelClass() : TclClass("NetworkModel") {}
	TclObject* create(int , const char*const* ) {
		return (new NetModel);
	}
} networkmodel_class;

NetModel::NetModel()
	: drawables_(0), animations_(0), queues_(0),
	  views_(0), nodes_(0), lans_(0), mon_count_(0), monitors_(NULL)
{
	int i;
	for (i = 0; i < EDGE_HASH_SIZE; ++i)
		hashtab_[i] = 0;

	/*XXX*/
	nymin_ = 1e6;
	nymax_ = -1e6;

	Paint *paint = Paint::instance();
	int p = paint->thin();
	nclass_ = 64;
	paint_ = new int[nclass_];
	for (i = 0; i < nclass_; ++i)
		paint_[i] = p;
}

NetModel::~NetModel()
{
}

void NetModel::update(double now)
{
	Animation *a, *n;
	for (a = animations_; a != 0; a = n) {
		n = a->next();
		a->update(now);
	}

	/*
	 * Draw all animations and drawables on display to reflect
	 * current time.
	 */
	now_ = now;
	for (NetView* p = views_; p != 0; p = p->next_) {
		p->draw();
	}
}

void NetModel::update(double now, Animation* a)
{
	// XXX Seems we should have a->update(now) here.
	a->update(now);
	for (NetView* p = views_; p != 0; p = p->next_)
		a->draw(p, now);
}

/* Reset all animations and queues to time 'now'. */
void NetModel::reset(double now)
{
	Animation* a;
	for (a = animations_; a != 0; a = a->next())
		a->reset(now);
	for (a = drawables_; a != 0; a = a->next())
		a->reset(now);
	for (Queue* q = queues_; q != 0; q = q->next_)
		q->reset(now);
}

/*
 * Draw this NetModel's animations and drawables (nodes, edges,
 * packets, queues).
 */
void NetModel::render(NetView* view)
{
	Animation *a;
	Monitor *m;
	for (a = drawables_; a != 0; a = a->next())
	  a->draw(view, now_);
	for (a = animations_; a != 0; a = a->next())
	  a->draw(view, now_);
	for(m=monitors_; m != NULL; m = m->next())
	  m->draw_monitor(view, nymin_, nymax_);
}

/* Return a pointer to the edge between 'src' and 'dst'. */
NetModel::EdgeHashNode* NetModel::lookupEdge(int src, int dst) const
{
	EdgeHashNode* h;
	for (h = hashtab_[ehash(src, dst)]; h != 0; h = h->next)
		if (h->src == src && h->dst == dst)
			break;
	return (h);
}

void NetModel::enterEdge(Edge* e)
{
	int src = e->src();
	int dst = e->dst();
	EdgeHashNode *h = lookupEdge(src, dst);
	if (h != 0) {
		/* XXX */
		fprintf(stderr, "nam: duplicate edge (%d,%d)\n", src, dst);
		exit(1);
	}
	h = new EdgeHashNode;
	h->src = src;
	h->dst = dst;
	h->queue = 0;
	h->edge = e;
	int k = ehash(src, dst);
	h->next = hashtab_[k];
	hashtab_[k] = h;
}

/* XXX Make this cheaper (i.e. cache it) */
void NetModel::BoundingBox(BBox& bb)
{
	/* ANSI C limits, from float.h */
	bb.xmin = bb.ymin = FLT_MAX;
	bb.xmax = bb.ymax = -FLT_MAX;
	
	for (Animation* a = drawables_; a != 0; a = a->next())
		a->merge(bb);
}


Animation* NetModel::inside(double now, float px, float py) const
{
	for (Animation* a = animations_; a != 0; a = a->next())
		if (a->inside(now, px, py))
			return (a);
	for (Animation* d = drawables_; d != 0; d = d->next())
		if (d->inside(now, px, py))
			return (d);

	return (0);
}

int NetModel::add_monitor(Animation *a)
{
  Monitor *m = new Monitor(mon_count_, a, node_size_);
  m->next(monitors_);
  monitors_=m;
  return mon_count_++;
}

int NetModel::monitor(double now, int monitor, char *result, int len)
{
  /*XXX should get rid of this search*/
  for(Monitor *m=monitors_; m!=NULL; m=m->next()) {
    if (m->monitor_number()==monitor) {
      m->update(now, result, len);
      if (strlen(result)==0)
	delete_monitor(m);
      return 0;
    }
  }
  result[0]='\0';
  return -1;
}

void NetModel::check_monitors(Animation *a)
{
  /*A new animation just got created.  Check to see if we should already*/
  /*have a monitor on it*/

  MonState *ms=a->monitor_state();
  if (ms==NULL) return;
  for(Monitor *m=monitors_; m!=NULL; m=m->next()) {
    if ((m->mon_state_!=NULL)&&(m->mon_state_->type=ms->type)) {
      switch (ms->type) {
      case MON_PACKET:
	if (m->mon_state_->pkt.id==ms->pkt.id) {
	  m->anim(a);
	  delete ms;
	  return;
	}
	break;
      case MON_ROUTE: 
	if ((m->mon_state_->route.src==ms->route.src)&&
	    (m->mon_state_->route.group==ms->route.group)) {
	  m->anim(a);
	  delete ms;
          return;
	}
	break;
      }
    }
  }
}

void NetModel::delete_monitor(int monitor)
{
  /*this version of delete_monitor is called from the GUI*/
  /*given a monitor, remove it from the model's monitor list*/
  Monitor *tm1, *tm2;
  if (monitors_==NULL)
    return;
  tm1=monitors_;
  tm2=monitors_;
  while ((tm1!=NULL)&&(tm1->monitor_number()!=monitor))
    {
      tm2=tm1;
      tm1=tm1->next();
    }
  if (tm1!=NULL)
    {
      tm2->next(tm1->next());
      if (tm1==monitors_)
	monitors_=tm1->next();
      if (tm1->anim()!=NULL)
	tm1->anim()->remove_monitor();
      delete tm1;
    }
}

void NetModel::delete_monitor(Monitor *m)
{
  /*given a monitor, remove it from a node's agent list*/
  Monitor *tm1, *tm2;
  if (monitors_==NULL)
    return;
  tm1=monitors_;
  tm2=monitors_;
  while ((tm1!=m)&&(tm1!=NULL))
    {
      tm2=tm1;
      tm1=tm1->next();
    }
  if (tm1!=NULL)
    {
      tm2->next(tm1->next());
      if (tm1==monitors_)
	monitors_=tm1->next();
      tm1->anim()->remove_monitor();
      delete tm1;
    }
}

void NetModel::delete_monitor(Animation *a)
{
  /*given a monitor, remove it from a node's agent list*/
  /*this version gets called when an animation deletes itself*/
  /*XXX animations sometimes get deleted when the packet changed link*/
  Monitor *tm1, *tm2;
  if (monitors_==NULL)
    return;
  tm1=monitors_;
  tm2=monitors_;
  while ((tm1!=NULL)&&(tm1->anim()!=a))
    {
      tm2=tm1;
      tm1=tm1->next();
    }
  if (tm1!=NULL)
    {
      tm2->next(tm1->next());
      if (tm1==monitors_)
	monitors_=tm1->next();
      delete tm1;
    }
}

void NetModel::saveState(double tim)
{
	State* state = State::instance();
	StateInfo min;
	min.time = 10e6;
	min.offset = 0;
	Animation *a, *n;
	StateInfo si;

	/*
	 * Update the animation list first to remove any unnecessary
	 * objects in the list.
	 */
	for (a = animations_; a != 0; a = n) {
		n = a->next();
		a->update(tim);
	}
	for (a = animations_; a != 0; a = n) {
		n = a->next();
		si = a->stateInfo();
		if (si.time < min.time)
			min = si;
	}
	if (min.offset)
		state->set(tim, min);
}

/* Trace event handler. */
void NetModel::handle(const TraceEvent& e, double now, int direction)
{
  QueueItem *q;
  EdgeHashNode *ehn;
  Edge* ep;
  Node *n;
  Packet *p;
  Route *r;
  Agent *a;
  Feature *f;
  float x, y;
  int pno;
  double txtime;
  if (e.time > State::instance()->next())
    saveState(e.time);
	
  switch (e.tt) 
    {
    case 'v':	/* 'variable' -- just execute it as a tcl command */
      Tcl::instance().eval((char *)(e.image + e.ve.str));
      break;
		
    case 'h':
      if (direction==BACKWARDS) break;
      ehn = lookupEdge(e.pe.src, e.pe.dst);
      if (ehn == 0 || (ep = ehn->edge) == 0)
	return;

      /*
       * If the current packet will arrive at its destination
       * at a later time, insert the arrival into the queue
       * of animations and set its paint id (id for X graphics
       * context.
       */
      txtime = ep->txtime(e.pe.pkt.size);
      if (e.time + txtime + ep->GetDelay() >= now) {
	/* XXX */
	Packet *p = new Packet(ep, e.pe.pkt, e.time, txtime,
			       e.offset);
	p->insert(&animations_);
	p->paint(paint_[e.pe.pkt.attr & 0xff]);
        check_monitors(p);
#ifdef DEBUG
	printf("packet %d sent from %d towards %d\n", e.pe.pkt.id, 
	       e.pe.src, e.pe.dst);
#endif
      }
      break;
    case 'a':
      n = lookupNode(e.ae.src);
      if (n == 0)
        return;
      if ((direction==FORWARDS)&&(e.ae.agent.expired==0)) {
	      a = n->find_agent((char *) e.ae.agent.name);
	      if (a == NULL) {
		      // will only create an agent if there isn't one already 
		      // there with the same name
		      a = new BoxAgent(e.ae.agent.name, n->size());
		      a->insert(&animations_);
		      n->add_agent(a);
		      placeAgent(a, n);
	      }
      } else {
	a=n->find_agent((char *)e.ae.agent.name);
	if (a != NULL) {
		n->delete_agent(a);
		delete a;
	}
      }
      break;
    case 'f':
      n = lookupNode(e.fe.src);
      if (n == 0)
        return;
      a = n->find_agent((char *)e.fe.feature.agent);
      if (a == 0)
	return;
      f = a->find_feature((char *)e.fe.feature.name);
      if (f == 0) {
	switch (e.fe.feature.type) {
	case 'v':
	  f = new VariableFeature(a, e.fe.feature.name);
	  break;
	case 'l':
	  f = new ListFeature(a, e.fe.feature.name);
	  break;
	case 's':
	case 'u':
	case 'd':
	  f = new TimerFeature(a, e.fe.feature.name);
	  break;
	}
	a->add_feature(f);
      }
      if (((direction==FORWARDS)&&(e.fe.feature.expired==0))||
	  ((direction==BACKWARDS)&&(strlen(e.fe.feature.oldvalue)>0))) {
	char *value;
	double time_set;
	if (direction==FORWARDS) {
	  value=(char *)e.fe.feature.value;
	  time_set=now;
	} else {
	  value=(char *)e.fe.feature.oldvalue;
	  /*XXX need an extra value here*/
	  time_set=0.0;
	}
	switch (e.fe.feature.type) {
	case 'v':
	case 'l':
	  f->set_feature(value);
	  break;
	case 's':
	  f->set_feature(atof(value), TIMER_STOPPED, time_set);
	  break;
	case 'u':
	  f->set_feature(atof(value), TIMER_UP, time_set);
	  break;
	case 'd':
	  f->set_feature(atof(value), TIMER_DOWN, time_set);
	  break;
	}
      } else {
	a->delete_feature(f);
	delete f;
      }
      break;
    case 'r':
      if (direction==FORWARDS) break;
      ehn = lookupEdge(e.pe.src, e.pe.dst);
      if (ehn == 0 || (ep = ehn->edge) == 0)
	return;

      /*
       * If the current packet will arrive at its destination
       * at a later time, insert the arrival into the queue
       * of animations and set its paint id (id for X graphics
       * context.
       */
      txtime = ep->txtime(e.pe.pkt.size);
      if (e.time - (txtime + ep->GetDelay()) <= now) {
	/* XXX */
	Packet *p = new Packet(ep, e.pe.pkt, 
			       e.time-(ep->GetDelay()+txtime), txtime,
			       e.offset);
	p->insert(&animations_);
	p->paint(paint_[e.pe.pkt.attr & 0xff]);
        check_monitors(p);
      }
      break;
    case '+':
      ehn = lookupEdge(e.pe.src, e.pe.dst);
      if (direction==FORWARDS)
	{
	  if (ehn != 0 && ehn->queue != 0) {
	    QueueItem *qi = new QueueItem(e.pe.pkt, e.time,
					  e.offset);
	    qi->paint(paint_[e.pe.pkt.attr & 0xff]);
	    ehn->queue->enque(qi,QUEUE_TAIL);
	    qi->insert(&animations_);
	    check_monitors(qi);
	  }
	}
      else
	{
	  if (ehn != 0 && ehn->queue != 0) {
	    q = ehn->queue->remove(e.pe.pkt);
	    delete q;
	  }
	}
      break;
		
    case '-':
      ehn = lookupEdge(e.pe.src, e.pe.dst);
      if (direction==FORWARDS)
        {
	  if (ehn != 0 && ehn->queue != 0) {
	    q = ehn->queue->remove(e.pe.pkt);
	    delete q;
	  }
	}
      else
	{
	  if (ehn != 0 && ehn->queue != 0) {
	    QueueItem *qi = new QueueItem(e.pe.pkt, e.time,
					  e.offset);
	    qi->paint(paint_[e.pe.pkt.attr & 0xff]);
	    ehn->queue->enque(qi,QUEUE_HEAD);
	    qi->insert(&animations_);
	    check_monitors(qi);
	  }
	}
      break;
    case 'l':
      /*link event*/
      ehn = lookupEdge(e.le.src, e.le.dst);
      if (ehn == 0)
	return;

      if (direction==FORWARDS)	{
        // Always save the color before the last DOWN event
	if (strncmp(e.le.link.state, "DOWN", 4)==0) {
	  ehn->edge->set_down("red");
	} else if (strncmp(e.le.link.state, "UP", 2)==0) {
	  /* XXX Assume an UP event always follows a DOWN event */
	  ehn->edge->set_up();
	} else if (strncmp(e.le.link.state, "COLOR", 5) == 0)
		ehn->edge->color((char *)e.le.link.color);
      } else {
	if (strncmp(e.le.link.state, "UP", 2)==0) {
	  ehn->edge->set_down("red");
	} else if (strncmp(e.le.link.state, "DOWN", 4)==0) {
	  ehn->edge->set_up();
	} else if (strncmp(e.le.link.state, "COLOR", 5) == 0)
		ehn->edge->color((char *)e.le.link.oldColor);
      }
      break;
    case 'n':
      /*node event*/
      n = lookupNode(e.ne.src);
      if (n == 0)
	return;
      if (direction==FORWARDS) {
	if (strncmp(e.ne.node.state, "DOWN", 4)==0) {
	  n->set_down("gray");
	} else if (strncmp(e.ne.node.state, "UP", 2)==0) {
	  n->set_up();
	} else if (strncmp(e.ne.node.state, "COLOR", 5) == 0) {
	      // normal color can be defined by user
	      // XXX: why color is opposite when direction is opposite?
	      n->color((char *)e.ne.node.color);
	}
      } else {
	if (strncmp(e.ne.node.state, "UP", 4)==0) {
	  n->set_down("gray");
	} else if (strncmp(e.ne.node.state, "DOWN", 2)==0) {
	  n->set_up();
	} else if (strncmp(e.ne.node.state, "COLOR", 5) == 0) {
	      n->color((char *)e.ne.node.oldColor);
	}
      }	
      break;
    case 'R':
      /*route event*/
      if (((e.re.route.expired==0)&&(direction==FORWARDS))||
	  ((e.re.route.expired==1)&&(direction==BACKWARDS)))
	{
	  /*this is a new route*/
	  n = lookupNode(e.re.src);
	  if (n == 0)
	    return;
	  ehn = lookupEdge(e.re.src, e.re.dst);
	  if (ehn == 0)
	    return;
	  int oif=1;
	  if (strncmp(e.re.route.mode, "iif", 3)==0)
	    oif=0;
	  r = new Route(n, ehn->edge, e.re.route.group, e.re.route.pktsrc, 
			e.re.route.neg, oif, e.re.route.timeout, now);
	  n->add_route(r);
	  placeRoute(r, n);
	  r->insert(&animations_);
	  r->paint(paint_[e.re.route.group & 0xff]);
	  check_monitors(r);
	}
      else
	{
	  /*an old route expired*/
	  n = lookupNode(e.re.src);
          if (n == 0)
            return;
	  ehn = lookupEdge(e.re.src, e.re.dst);
          if (ehn == 0)
            return;
	  int oif=1;
	  if (strncmp(e.re.route.mode, "iif", 3)==0)
	    oif=0;
	  r = n->find_route(ehn->edge, e.re.route.group, e.re.route.pktsrc,
			    oif);
	  if (r==0)
	    {
	      fprintf(stderr, "nam: attempt to delete non-existent route\n");
	      abort();
	    }
	  n->delete_route(r);
	  delete r;
	}
      break;
    case 'd':
      /* Get packet to drop. */
      if (direction==FORWARDS)
	{
#ifdef DEBUG
	  printf("drop on %d -> %d\n", e.pe.src, e.pe.dst);
#endif
	  ehn = lookupEdge(e.pe.src, e.pe.dst);
	  if (ehn == 0)
	    return;
	  q=0;
	  if (ehn->queue!=0)
	    {
	      /*if there's a queue, try removing it from the queue first*/
	      /*queue drops are more common than link drops...*/
	      q = ehn->queue->remove(e.pe.pkt);
	    }
	  if (q == 0) {
	    /*perhaps it's a link packet drop*/
	    p=lookupPacket(e.pe.src, e.pe.dst, e.pe.pkt.id);
	    if (p!=NULL)
	      {
		/*it was a link packet drop*/
#ifdef DEBUG
		printf("packet %d dropped on link %d->%d\n", e.pe.pkt.id, e.pe.src, e.pe.dst);
#endif
		ehn->edge->DeletePacket(p);
		p->position(x, y, now);
		pno = p->paint();
		delete p;
	      }
	    else 
	      {
		/*probably it was a queue drop, but we're not displaying */
		/*that queue*/
		return;
	      }
	  } else {
#ifdef DEBUG
	    printf("packet %d dropped from queue\n", e.pe.pkt.id);
#endif
	    q->position(x, y);
	    pno = q->paint();
	    delete q;
	  }
	  /*
	   * Compute the point at which the dropped packet disappears.
	   * Let's just make this sufficiently far below the lowest
	   * thing on the screen.
	   *
	   * Watch out for topologies that have all their nodes lined
	   * up horizontally. In this case, nymin_ == nymax_ == 0.
	   * Set the bottom to -0.028. This was chosen empirically.
	   * The nam display was set to the maximum size and the lowest
	   * position on the display was around -0.028.
	   */
	  float bot;
	  if (nymin_ == nymax_ && nymin_ == 0.)
	    bot = -0.03;
	  else
	    bot = nymin_ - (nymax_ - nymin_);
	  
	  Drop *d = new Drop(x, y, bot, ehn->edge->PacketHeight(),
			     e.time, e.offset, e.pe.pkt);
	  d->paint(pno);
	  d->insert(&animations_);
	  check_monitors(d);
	  break;
	} else {
	  /*direction is BACKWARDS - need to create the packet*/
	}
    }
}

void NetModel::addView(NetView* p)
{
	p->next_ = views_;
	views_ = p;
}

Node *NetModel::lookupNode(int nn) const
{
	for (Node* n = nodes_; n != 0; n = n->next_)
		if (n->num() == nn)
			return (n);
	/* XXX Do not exit, we need normal return. */
	//fprintf(stderr, "nam: no such node %d\n", nn);
	//exit(1);
	return (0);// make visual c++ happy
}

void NetModel::removeNode(Node *n) 
{
	Node *p, *q;
	// Remove node n from nodes_ list, then delete it
	for (p = nodes_; p != 0; q = p, p = p->next_)
		if (p == n)
			break;
	if (p == nodes_)
		nodes_ = p->next_;
	else
		q->next_ = p->next_;
	delete p;
}

Lan *NetModel::lookupLan(int nn) const
{
	for (Lan* l = lans_; l != 0; l = l->next_)
		if (l->num() == nn)
			return (l);
	/* XXX */
	fprintf(stderr, "nam: no such lan %d\n", nn);
	exit(1);
	return (0);// make visual c++ happy
}

Packet *NetModel::lookupPacket(int src, int dst, int id) const
{
  EdgeHashNode *h = lookupEdge(src, dst);
  if (h == 0)
    return NULL;
  int ctr=0;
  for (Packet *p=h->edge->packets(); p!=NULL; p=p->next())
    {
#define PARANOID
#ifdef PARANOID
      ctr++;
      if (ctr>h->edge->no_of_packets()) abort();
#endif
      if (p->id() == id)
	return p;
    }
  /*have to fail silent or we can't cope with link drops when doing settime*/
  return 0;
}

int NetModel::command(int argc, const char *const *argv)
{
	Tcl& tcl = Tcl::instance();
	
	if (argc == 2) {
		if (strcmp(argv[1], "layout") == 0) {
			/*
			 * <net> layout
			 * Compute reasonable defaults for missing node or edge
			 * sizes based on the maximum link delay. Lay out the
			 * nodes and edges as specified in the layout file.
			 */
			scale_estimate();
			layout();
			return (TCL_OK);
		}
	} else if (argc == 3) {
		if (strcmp(argv[1], "view") == 0) {
			/*
			 * <net> view <viewName>
			 * Create the window for the network layout/topology.
			 */
			NetView *v = new NetView(argv[2], this);
			v->next_ = views_;
			views_ = v;
			return (TCL_OK);
		}
	} else if (argc >= 4 && strcmp(argv[1], "node") == 0) {
		/*
		 * <net> node <name> <shape> <color> [<size>]
		 * Create a node using the specified name
		 * and the default size and insert it into this
		 * NetModel's list of drawables.
		 */
		double size = defsize;
		if (argc > 5)
			size = atof(argv[5]);
		/* 
		 * Check whether this node is already there. If so, 
		 * Delete it first
		 */
		Node *n = lookupNode(atoi(argv[2]));
		if (n != NULL) {
			fprintf(stderr, "duplicate node %s.\n", argv[2]);
			removeNode(n);
		}
		if (!strcmp(argv[3], "circle"))
			n = new CircleNode(argv[2], size);
		else if (!strcmp(argv[3], "box"))
			n = new BoxNode(argv[2], size);
		else
			n = new HexagonNode(argv[2], size);
		n->init_color((char *)argv[4]);
		n->next_ = nodes_;
		nodes_ = n;
		n->insert(&drawables_);
		return (TCL_OK);
	} else if (argc == 4) {
		if (strcmp(argv[1], "color") == 0) {
			/*
			 * <net> color <packetClass> <colorName>
			 * Display packets of the specified class using
			 * the specified color.
			 */
			int c = atoi(argv[2]);
			if ((u_int)c > 1024) {
				tcl.resultf("%s color: class %d out of range",
					    argv[0], c);
				return (TCL_ERROR);
			}
			Paint *paint = Paint::instance();
			if (c > nclass_) {
				int n, i;
				for (n = nclass_; n < c; n <<= 1);
				int *p = new int[n];
				for (i = 0; i < nclass_; ++i)
					p[i] = paint_[i];
				delete paint_;
				paint_ = p;
				nclass_ = n;
				int pno = paint->thin();
				for (; i < n; ++i)
					paint_[i] = pno;
			}
			int pno = paint->lookup(Tk_GetUid((char *)argv[3]), 1);
			if (pno < 0) {
			  fprintf(stderr,"%s color: no such color: %s\n",
				  argv[0], argv[3]);
			  pno = paint->lookup(Tk_GetUid("black"),1);
			  if (pno < 0) {
			    tcl.resultf("%s no black! - bailing");
			    return (TCL_ERROR);
			  }
			}
			paint_[c] = pno;
			return (TCL_OK);
		}
		if (strcmp(argv[1], "ncolor") == 0) {
			/*
			 * <net> ncolor <node> <colorName>
			 * set color of node to the specified color.
			 */
			Paint *paint = Paint::instance();
			Node *n = lookupNode(atoi(argv[2]));
			if (n == NULL) {
				fprintf(stderr, "No such node %s\n", argv[2]);
				exit(1);
			}
			int pno = paint->lookup(Tk_GetUid((char *)argv[3]), 3);
			if (pno < 0) {
			  fprintf(stderr,"%s ncolor: no such color: %s\n",
				  argv[0], argv[3]);
			  int pno = paint->lookup(Tk_GetUid("black"),1);
			  if (pno < 0) {
			    tcl.resultf("%s no black! - bailing");
			    return (TCL_ERROR);
			  }
			}
			n->paint(pno);
			return (TCL_OK);
		}
	} else if (argc == 5) {
		if (strcmp(argv[1], "queue") == 0) {
			/*
			 * <net> queue <src> <dst> <angle>
			 * Create a queue for the edge from 'src' to 'dst'.
			 * Add it to this NetModel's queue list.
			 * Display the queue at the specified angle from
			 * the edge to which it belongs.
			 */
			int src = atoi(argv[2]);
			int dst = atoi(argv[3]);
			EdgeHashNode *h = lookupEdge(src, dst);
			if (h == 0) {
				tcl.resultf("%s queue: no such edge (%d,%d)",
					    argv[0], src, dst);
				return (TCL_ERROR);
			}
			/* XXX can we assume no duplicate queues? */
			double angle = atof(argv[4]);
			Edge *e = h->edge;
			angle += e->angle();
			Queue *q = new Queue(angle);
			h->queue = q;
			q->next_ = queues_;
			queues_ = q;
			return (TCL_OK);
		}
		if (strcmp(argv[1], "ecolor") == 0) {
			/*
			 * <net> ecolor <src> <dst> <colorName>
			 * set color of edge to the specified color.
			 */
			Paint *paint = Paint::instance();
			EdgeHashNode* h = lookupEdge(atoi(argv[2]), atoi(argv[3]));
			if (h == 0) {
				tcl.resultf("%s ecolor: no such edge (%s,%s)",
					    argv[0], argv[2], argv[3]);
				return (TCL_ERROR);
			}
			int pno = paint->lookup(Tk_GetUid((char *)argv[4]), 3);
			if (pno < 0) {
			  fprintf(stderr,"%s ncolor: no such color: %s\n",
				  argv[0], argv[3]);
			  pno = paint->lookup(Tk_GetUid("black"),1);
			  if (pno < 0) {
			    tcl.resultf("%s no black! - bailing");
			    return (TCL_ERROR);
			  }
			}
			h->edge->paint(pno);
			return (TCL_OK);
		}
		if (strcmp(argv[1], "lanlink") == 0) {
			/*
			 * <net> link <src> <lan> <angle>
			 * Create a link/edge between the specified source
			 * and a lan.
			 */
			Node *src = lookupNode(atoi(argv[2]));
			Lan *lan = lookupLan(atoi(argv[3]));
			double angle = atof(argv[4]);
			double bw=lan->bw();
			double delay=(lan->delay())/2.0;
			Edge *e1 = new Edge(src, lan->virtual_node(), defsize,
					    bw, delay, angle+1);
			e1->init_color("black");
			enterEdge(e1);
			e1->insert(&drawables_);
			src->add_link(e1);

			Edge *e2 = new Edge(lan->virtual_node(), src, defsize,
					   bw, delay, angle);
			e2->init_color("black");
			enterEdge(e2);
			e2->insert(&drawables_);
			lan->add_link(e2);
			return (TCL_OK);
		}
	} else if (argc == 6) {
		if (strcmp(argv[1], "lan") == 0) {
			/*
			 * <net> lan <name> <bandwidth> <delay> <angle>
			 * Create a link/edge between the specified source
			 * and destination. Add it to this NetModel's list
			 * of drawables and to the source's list of links.
			 */
			double bw = atof(argv[3]);
			double delay = atof(argv[4]);
			double angle = atof(argv[5]);
			Lan *l = new Lan(argv[2], defsize, bw, delay, angle);
			l->next_=lans_;
			lans_=l;
			l->insert(&drawables_);
			Node *n=l->virtual_node();
			n->next_ = nodes_;
			nodes_ = n;
			return (TCL_OK);
		}
	} else if (argc == 7) {
		if (strcmp(argv[1], "link") == 0) {
			/*
			 * <net> link <src> <dst> <bandwidth> <delay> <angle>
			 * Create a link/edge between the specified source
			 * and destination. Add it to this NetModel's list
			 * of drawables and to the source's list of links.
			 */
			Node *src = lookupNode(atoi(argv[2]));
			Node *dst = lookupNode(atoi(argv[3]));
			double bw = atof(argv[4]);
			double delay = atof(argv[5]);
			double angle = atof(argv[6]);
			Edge *e = new Edge(src, dst, defsize,
					   bw, delay, angle);
			e->init_color("black");
			enterEdge(e);
			e->insert(&drawables_);
			src->add_link(e);
			tcl.resultf("%g", delay);
			return (TCL_OK);
		}
	}
	/* If no NetModel commands matched, try the Object commands. */
	return (TclObject::command(argc, argv));
}

void NetModel::placeEdge(Edge* e, Node* src) const
{
	if (e->marked() == 0) {
		double nsin, ncos;
		Node *dst = e->neighbor();
		SINCOSPI(e->angle(), &nsin, &ncos);
		double x0 = src->x() + src->size() * ncos * .75;
		double y0 = src->y() + src->size() * nsin * .75;
		double x1 = dst->x() - dst->size() * ncos * .75;
		double y1 = dst->y() - dst->size() * nsin * .75;
		e->place(x0, y0, x1, y1);
		
		/* Place the queue here too.  */
		EdgeHashNode *h = lookupEdge(e->src(), e->dst());
		if (h->queue != 0)
			h->queue->place(e->size(), e->x0(), e->y0());

		e->mark();
	}
}

void NetModel::placeRoute(Route* r, Node* src) const
{
	if (r->marked() == 0) {
		double nsin, ncos;
		//Unused: Node *dst = r->edge()->neighbor();
		SINCOSPI(r->angle(), &nsin, &ncos);
		double x0 = src->x() + src->size() * ncos * .75;
		double y0 = src->y() + src->size() * nsin * .75;
		r->place(x0, y0);
		r->mark(1);
	}
}

void NetModel::placeAgent(Agent *a, Node *src) const
{
	double x0, y0, nsin, ncos;
	if (a->marked() == 0) {
		if (a->angle()==NO_ANGLE) {
			if (a->edge()==NULL) {
			  /* determine where to put the label so it
			     won't overlap a link (if possible) */
			        int choices[8], i, ix;
				for(i=0;i<8;i++)
				  choices[i]=1;
				Agent *agents = src->agents();
                                while (agents!=NULL) {
				  ix=int(agents->angle()*4);
				  choices[ix]=0;
				  agents=agents->next_;
				}
				Edge *links = src->links();
                                while (links!=NULL) {
				  ix=int(links->angle()*4);
				  choices[ix]=0;
				  links=links->next_;
				}
				if (choices[0]==1)
				  a->angle(0);
				else if (choices[4]==1)
				  a->angle(1.0);
				else if ((choices[1]==1)&&(choices[2]==1))
				  a->angle(0.5);
				else if ((choices[6]==1)&&(choices[7]==1))
				  a->angle(1.5);
				else if (choices[1]==1)
				  a->angle(0.25);
				else
				  a->angle(1.75);
			} else {
				a->angle(a->edge()->angle()+0.25);
			}
		}
		SINCOSPI(a->angle(), &nsin, &ncos);
		x0 = src->x() + src->size() * ncos * .75;
		y0 = src->y() + src->size() * nsin * .75;
		a->place(x0, y0);
		a->mark(1);
	}
}

void NetModel::layout()
{
	/* If there is no fixed node, anchor the first one entered at (0,0). */
	Node *n;

	for (n = nodes_; n != 0; n = n->next_)
		n->mark(0);
	if (nodes_)
		nodes_->place(0., 0.);

	int did_something;
	do {
		did_something = 0;
		for (n = nodes_; n != 0; n = n->next_)
			did_something |= traverse(n);
	} while (did_something);

	for (n = nodes_; n != 0; n = n->next_)
		for (Edge* e = n->links(); e != 0; e = e->next_)
			placeEdge(e, n);
}

void NetModel::move(double& x, double& y, double angle, double d) const
{
	double ncos, nsin;
	SINCOSPI(angle, &nsin, &ncos);
	x += d * ncos;
	y += d * nsin;
}

/*
 * Traverse node n's neighbors and place them based on the
 * delay of their links to n.  The two branches of the if..else
 * are to handle unidirectional links -- we place ourselves if
 * we haven't been placed & our downstream neighbor has.
 */
int NetModel::traverse(Node* n)
{
	int did_something = 0;
	for (Edge* e = n->links(); e != 0; e = e->next_) {
		Node *neighbor = e->neighbor();
		double d = e->delay() + (n->size() + neighbor->size()) * 0.75;
		if (n->marked() && !neighbor->marked()) {
			double x = n->x();
			double y = n->y();
			move(x, y, e->angle(), d);
			neighbor->place(x, y);
			did_something |= traverse(neighbor);
			if (nymax_ < y)
				nymax_ = y;
			if (nymin_ > y)
				nymin_ = y;
		} else if (!n->marked() && neighbor->marked()) {
			double x = neighbor->x();
			double y = neighbor->y();
			move(x, y, e->angle(), -d);
			n->place(x, y);
			did_something = 1;
		}
	}
	return (did_something);
}

/*
 * Compute reasonable defaults for missing node or edge sizes
 * based on the maximum link delay.
 */
void NetModel::scale_estimate()
{
	/* Determine the maximum link delay. */
	double emax = 0.;
	Node *n;
	for (n = nodes_; n != 0; n = n->next_) {
		for (Edge* e = n->links(); e != 0; e = e->next_)
			if (e->delay() > emax)
				emax = e->delay();
	}
	
	/*
	 * Check for missing node or edge sizes. If any are found,
	 * compute a reasonable default based on the maximum edge
	 * dimension.
	 */
	for (n = nodes_; n != 0; n = n->next_) {
		if (n->size() <= 0.)
			n->size(.1 * emax);
		for (Edge* e = n->links(); e != 0; e = e->next_)
			if (e->size() <= 0.)
				e->size(.03 * emax);
	}
	/*store this because we need it for monitors*/
	node_size_=.1 * emax;
}
